#!/usr/bin/perl

use strict;
use PerlIO;               # When I added this line, and used [ open MEDCITATION ">:utf8", "...", ], I no longer got "Wide character in print" warnings.  Encodes output as UTF8.
use Encode;               # Used in conjunction with DBM Filters so that we can tie hashes with UTF8 Keys.
use XML::Twig;
use DB_File;
use DBI;
use IO::Zlib;
use File::Basename;

my $TEMP_RESULTS_DIR = ".";
my $XML_FILE = "test.xml";

my $time = time;
#------------------------------------------------
# variables
#------------------------------------------------
my %stopwords = ();            # simple stopword list
my %word_pmid_counts =();         # hash for collecting PMID:sum( log1.6( COUNT^2 ):COUNT for each word in this Medline file
my %word_pmid_counts_save = ();         # This is the hash we're going to save to the hard disk
my %pmid_count_sums = ();
my %pmid_count_sums_save = ();

my %log_scores = ();            # Going to cache log(x +.6)/log(.6) x = 1..10
my %this_pmid = ();            # Make a small hash for each PMID the XML parser reads
my $pmid_count;
my @deleted_citations;            # each MEDLINE file has a list of deleted citations at the end
my @authors;               # each PMID gets it own list of authors
my $is_revised;               # each PMID gets a boolean indicating it's revised or not
my $date_created_revised;         # each PMID was either created or revised on the date this XML file was sent
my %year_count = ();            # this collects the number of publications per year //Chris


#-------------------------------------------------------------
# init publication hash with years 1800 to 2100 at 0
#------------------------------------------------------------
for(my $l = 0; $l <= 300; $l++)
{
   $year_count{$l} = 0;
}

# For efficiency, build a table of log scores for counts 1..10,
# to save recalculating them over and over for millions of words
#----------------------------------------------------------------------
my $log_1dot6   = log(1.6);                     # constant
for (1..10)
{
        $log_scores{$_} = log( $_ + .6)/$log_1dot6;
}


my $twig = XML::Twig->new( twig_handlers =>
            { 
               MedlineCitation => \&parse_citation,
               PMID => \&parse_pmid,
               AbstractText => \&parse_abstract,
                 ArticleTitle => \&parse_title,
               Author=> \&parse_authors,
               Affiliation => \&parse_affiliation,
               DateRevised => \&parse_date_revised,
               DateCreated => \&parse_date_created,
               Journal => \&parse_journal,
               MedlineTA => \&parse_journal_title,
               MedlinePgn => \&parse_page_number,
               CopyrightInformation => \&parse_copyright,
               DeleteCitation => \&parse_deleted_citation,
               PublicationTypeList => \&parse_pub_type,
            },
            
             ignore_elts =>{ CommentsCorrections => 1},      # ignore <PMID> tags inside <CommentsCorrections> tags
         );


%word_pmid_counts = ();

$twig->parsefile( $XML_FILE );

print "Parsing complete. \n";


###################################
# SUBS
###################################


#---------------------------------------------------
# Called by parse_citation after a each XML PMID
# citation is called.  process_pmid pulls the keywords
# out of the abstract and title, strips punctuation
# and writes them to a BerkeleyDB.
#----------------------------------------------------
sub process_pmid
{
   my %keywords = ();
   my $text;
        if( $this_pmid{abstract} )                              # grab all the words in the abstract
        {
                $text = lc $this_pmid{abstract};
        }
   $text .= " " unless not defined $text;
   $text .= lc $this_pmid{title};
   $text .= " ".lc $this_pmid{authors};

   $text =~ s/[\{\}'\".,=+):;\(|*&^\#!@%\$\/_?]/ /g;       # get rid of punctuation marks.  Don't remove dashes...
        $text =~ s/\s\d+\s/ /g;                                # remove numbers, if they stand alone
        $text =~ s/^\d+\s/ /g;
        $text =~ s/\s\d+$/ /g;
        my @all_words = split /\s+/,  $text;
        my $sum_log_square_count_this_pmid = 0;# initialize weighted, dicounted word vector dot product for cosine function
   my $documentHz = 0;
   $keywords{$_}++ for @all_words;
   delete @keywords{keys %stopwords};
   $documentHz += $_ * $_ for map($log_scores{$_} ||= log($_ + .6)/$log_1dot6, values %keywords);
   $sum_log_square_count_this_pmid = $documentHz;
   while( my($key,$value) = each %keywords)
   {

      print WORDCOUNTS $key . "\t" . $this_pmid{pmid} . "\t" . $documentHz . "\t" . $value . "\n";

#      if( defined $word_pmid_counts{$key} )
#      {
#         $word_pmid_counts{$key} .= pack('IfS', $this_pmid{pmid}, $documentHz, $value);
#
#      }
#      else
#      {
#         $word_pmid_counts{$key} = pack('IfS', $this_pmid{pmid}, $documentHz, $value);
#      }

   }
   $pmid_count++;
}
#--------------------------------
# Called by the call to twig->parsefile in main
#-------------------------------
sub parse_citation
{
   my ( $twig, $citation ) = @_;
   my $citation_info = $this_pmid{ journal_name }." ".$this_pmid{ partial_citation_info }.$this_pmid{ pages }.".";
   $this_pmid{ affiliation } = "Affiliation not provided" unless defined $this_pmid{ affiliation };
   $this_pmid{ publisher } = "NULL" unless defined $this_pmid{ publisher };
   #------------------------------------------------------
   # if date_revised is defined, then this XML file represents a revised entry.
   # if it's not, then it's a new MEDLINE entry.
   #------------------------------------------------------
   if ( $date_created_revised = $this_pmid{ date_revised } )
   {
      $is_revised = 1;
   }
   else
   {
      $date_created_revised = $this_pmid{ date_created };
      $is_revised = 0;
   }
   #-------------------------------------------
   # Write this record to MySQL loader file
   #-------------------------------------------
#   print MEDCITATION "$date_created_revised\t$is_revised\t$this_pmid{ title }\t$this_pmid{ journal_issn }\t$this_pmid{ publisher }\t$this_pmid{ abstract }\t$this_pmid{ pmid }\t$citation_info\t$this_pmid{ affiliation }\t$this_pmid{ pub_year }\n" or warn "$this_pmid{ pmid }: $!\n";
   while( my $author = shift @authors )
   {
      print AUTHORS "$this_pmid{ pmid }\t$date_created_revised\t$author\n";
   }
   $twig->purge;            # When you've loaded this PMID element into %this_pmid, delete this twig.
   &process_pmid();         # process this pmid
   %this_pmid = ();         # Reset the hash.
}

sub parse_pmid
{
   my ( $twig, $PMID) = @_;
   $this_pmid{ pmid } =  $PMID->xml_text;
   print STDERR "parse_pmid:  PMID=" . $this_pmid{ pmid } . "\n";

   chomp $this_pmid{ pmid };
}
#--------------------------------------------
# Because we're running Perl 5.8+, we should
# already recognize UTF8 character encoding.
# So no filter on input.
#-------------------------------------------
sub parse_abstract
{
   my ( $twig, $abstract ) = @_;
   $this_pmid{ abstract } =  ( $abstract->xml_text || "");
}

sub parse_title
{
   my ( $twig, $title) = @_;
   $this_pmid{ title } = ( $title->xml_text || "" );
   chomp $this_pmid{ title };
}

sub parse_authors
{
   my ( $twig, $authorlist ) = @_;
   return if  $authorlist->first_child_text( 'CollectiveName');
   my $last_name = $authorlist->first_child_text( 'LastName');
   my $first_name = $authorlist->first_child_text( 'ForeName');
   my $author = $first_name." ".$last_name;
   $this_pmid{ authors } .= " ".$last_name;
   push @authors, $author;
}

sub parse_date_revised
{
   my ( $twig, $date_revised ) = @_;
   my $date_string = $date_revised->first_child_text( 'Year' );
   $date_string .= "-".$date_revised->first_child_text( 'Month' );
   $date_string .= "-".$date_revised->first_child_text( 'Day' );
   $this_pmid{ date_revised } =  $date_string;
   $this_pmid{ date_revised } ||= "";
   chomp  $this_pmid{ date_revised };
}

sub parse_date_created
{
   my ( $twig, $date_created) = @_;
   my $date_string = $date_created->first_child_text( 'Year' );
   $date_string .= "-".$date_created->first_child_text( 'Month' );
   $date_string .= "-".$date_created->first_child_text( 'Day' );
   $this_pmid{ date_created } =  $date_string;
   chomp $this_pmid{ date_created };
}

sub parse_journal
{
   my ( $twig, $journal ) = @_;
   my $ISSN = $journal->first_child_text( 'ISSN' );
   chomp $ISSN;
   $this_pmid{ journal_issn } = $ISSN;
   # Commented this line so I could break publication year out as a separate field.  Replaced with 3 subsequent lines.
   #my $citation_info = $journal->first_child( 'JournalIssue' )->first_child('PubDate')->first_child_text( 'Year' );
   my $pub_year = $journal->first_child( 'JournalIssue' )->first_child('PubDate')->first_child_text( 'Year' );
   $this_pmid{ pub_year } = $pub_year;
   my $citation_info = $pub_year;
   $citation_info .= " ".$journal->first_child( 'JournalIssue' )->first_child('PubDate')->first_child_text( 'Month' );
   my $temp_year = $journal->first_child( 'JournalIssue' )->first_child('PubDate')->first_child_text( 'MedlineDate' );
   if($pub_year == "")
   {
      $pub_year = $temp_year;
   }
   $citation_info .= $temp_year;
   $year_count{($pub_year - 1800)} = 1 + $year_count{($pub_year - 1800)};
   $citation_info .= "; ".$journal->first_child( 'JournalIssue' )->first_child_text('Volume'); # <Journal><JournalIssue><Volume>1</Volume><Issue>1</Issue></JournalIssue></Journal>.... sorry this is ugly.  ->child(1, 'Volume') didn't seem to work for some reason
   $citation_info .= "(".$journal->first_child( 'JournalIssue' )->first_child_text( 'Issue' ).")";
   $this_pmid{ partial_citation_info } = $citation_info;   
   chomp $this_pmid{ partial_citation_info };
}

sub parse_journal_title
{
   my ( $twig, $medlineTA ) = @_;
   $this_pmid{ journal_name } = $medlineTA->xml_text;
   chomp $this_pmid{ journal_name };
}

sub parse_page_number
{
   my ( $twig, $MedlinePgn ) = @_;
    $this_pmid{ pages } = $MedlinePgn->xml_text;
   chomp  $this_pmid{ pages };
}

sub parse_affiliation
{
   my ( $twig, $affiliation ) = @_;
   $this_pmid{ affiliation } = $affiliation->xml_text;
   chomp $this_pmid{ affiliation };
}

sub parse_deleted_citation
{
   my ($twig, $deleted ) = @_;
   # NOTE: the version of XML::Twig running on the individual storm nodes
   # is different from the one on the master node, and doesn't have the
   # children_text function.  Took me HOURS to realize that!
   #@deleted_citations = $deleted->children_text('PMID');
   foreach ( $deleted->children('PMID') )
   {
       push @deleted_citations, $_->text;
   }
}

sub parse_copyright
{
   my ($twig, $copyright) = @_;
   $this_pmid{ publisher} = $copyright->xml_text || "NULL";
}

sub parse_pub_type
{
   # NOTE: The PublicationTypeList xml tag implies multiple pub types.
   # I'm just catting them together, and will grep this
   # string for particular types when it's time to
   # search this database...kindof hackish.
   my ($twig, $pub_type) = @_;
   my $pub_string = "";
   foreach ( $pub_type->children('PublicationType') )
   {
         $pub_string .= $_->text;
   }
   $this_pmid{ pub_type} = $pub_string;
}
